/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Author:
 *   Alvaro del Castillo (acs@lambdaux.com)
 *
 * Copyright 2004, LambdaUX
 * 
 * First try based in the File contacts backend by Nat Friedman
 */

#ifdef HAVE_CONFIG_H
#include <config.h> 
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>

#include <libgnome/gnome-i18n.h>
#include <libedataserver/e-url.h> 
#include <libebook/e-contact.h>
#include <libebook/e-address-western.h>

#include <libedata-book/e-book-backend-sexp.h>
#include <libedata-book/e-book-backend-summary.h>
#include <libedata-book/e-data-book.h>
#include <libedata-book/e-data-book-view.h>
#include "e-book-backend-ogo.h"
#include <utils/ogo-connection.h>
#include <utils/ogo-book-backend-cache.h>
#include <utils/e2k-properties.h>
#include <utils/e2k-propnames.h>
#include <utils/e2k-utils.h>

#define d(x) x

#define SUMMARY_FLUSH_TIMEOUT 5000

static EBookBackendClass *parent_class;

struct _EBookBackendOGoPrivate {
	OGoConnection       *conn;
	
	OGoBookBackendCache *cache;
	
	gchar               *uri; 
	gchar               *item_path;
	gchar               *use_ssl;
	gboolean             only_if_exists;
	int                  mode;
};

static GPtrArray *field_names_array;
static const char **field_names;
static int n_field_names;

#define OGO_ETAG_PROPERTY "X-OGO-ETAG"
#define OGO_HREF_PROPERTY "X-OGO-HREF"

static void get_ogo_info (EContact *contact, gchar **href, gchar **etag);
static const char *contact_name (EContact *contact);
static E2kProperties *props_from_contact (EBookBackendOGo *be,
					  EContact *contact,
					  EContact *old_contact);

/* This function takes a GroupDAV uri and turns it into
 * one readable by the Evolution Connector support in the server.
 */
static gchar *
get_e2k_uri (const gchar *uri)
{
	gchar **tmpv;
	EUri *parsed_uri;
	GString *new_uri;
	int i;
	gchar *e2k_uri, *tmp;
	
	parsed_uri = e_uri_new (uri);
	tmpv = g_strsplit (parsed_uri->path, "/", 0);
	
	g_assert (strcmp (tmpv[0], "") == 0);
	g_assert (strcmp (tmpv[1], "zidestore") == 0);
	g_assert (strcmp (tmpv[2], "dav") == 0);
	
	new_uri = g_string_new ("/zidestore");
 	for (i = 4; tmpv[i] != NULL; i++) {
		g_string_append_printf (new_uri, "/%s", tmpv[i]);
	}

	g_free (parsed_uri->path);
	parsed_uri->path = g_string_free (new_uri, FALSE);
	g_strfreev (tmpv);

	e2k_uri = e_uri_to_string (parsed_uri, TRUE);
	e_uri_free (parsed_uri);
	
	return e2k_uri;
}

static gboolean
test_name (OGoConnection   *conn, 
	   const char      *name,
	   gpointer         user_data)
{
	EBookBackendOGo *ogo;
	
	g_return_if_fail (E_IS_BOOK_BACKEND_OGO (user_data));
	ogo = E_BOOK_BACKEND_OGO (user_data);
	
	return !ogo_book_backend_cache_get_contact (ogo->priv->cache, name);
}

static void
e_book_backend_ogo_create_contact (EBookBackend *backend,
				   EDataBook    *book,
				   guint32       opid,
				   const char   *vcard)
{
	EBookBackendOGo     *ogo = E_BOOK_BACKEND_OGO (backend);
	EContact            *contact;
	E2kProperties       *props;
	const gchar         *name;
	gchar               *e2k_href, *href;
	gchar               *location;
	OGoConnectionStatus  status;
	
	contact = e_contact_new_from_vcard (vcard);
	
	props = props_from_contact (ogo, contact, NULL); 

	/* figure out the right uri to be using */
	name = contact_name (contact);
	if (!name)
		name = "No Subject";

	href = ogo_connection_get_item_uri (ogo->priv->conn, ogo->priv->item_path);
	e2k_href = get_e2k_uri (href);
	g_free (href);
	
	status = ogo_connection_proppatch_new (ogo->priv->conn, e2k_href, name,
					       test_name, ogo,
					       props, &location, NULL);
	g_free (e2k_href);
	g_warning ("%s: implement me!", __PRETTY_FUNCTION__);
}

static void
e_book_backend_ogo_remove_contacts (EBookBackend *backend,
				    EDataBook    *book,
				    guint32       opid,
				    GList        *id_list)
{
	EBookBackendOGo *ogo = E_BOOK_BACKEND_OGO (backend);
	GList *list;
	GList *deleted_ids = NULL;
	
	if (!ogo->priv->conn) {
		e_data_book_respond_remove_contacts (book, opid, GNOME_Evolution_Addressbook_AuthenticationRequired, NULL);
		return;
	}

	for (list = id_list; list != NULL; list = list->next) {
		EContact *contact;
		gchar *href, *etag;
		OGoConnectionStatus status;
		
		contact = ogo_book_backend_cache_get_contact (ogo->priv->cache, list->data);

		if (!contact)
			continue;

		get_ogo_info (contact, &href, &etag);
		g_object_unref (contact);
		
		status = ogo_connection_delete_item (ogo->priv->conn, href, etag);
		g_free (href);
		g_free (etag);

		
		if (status != OGO_CONNECTION_STATUS_OK) {
			g_warning ("Failed to remove contact: %d", status);
			continue;
		}

		ogo_book_backend_cache_remove_contact (ogo->priv->cache, list->data);
		deleted_ids = g_list_append (deleted_ids, list->data);
	}

	e_data_book_respond_remove_contacts (book, opid,
					     GNOME_Evolution_Addressbook_Success,  deleted_ids);
}

typedef struct PropMapping PropMapping;

struct PropMapping {
	EContactField field;
	char *prop_name;
	char *e_book_field;
	int flags;
#define FLAG_UNLIKEABLE 0x001  /* can't use LIKE with it */
#define FLAG_COMPOSITE  0x002  /* read-only fields that can be written
                                 to only by specifying the values of
                                 individual parts (as other fields) */
#define FLAG_PUT        0x020  /* requires a PUT request */
#define FLAG_EMAIL      0x100  /* email field, so we know to invoke our magic email address handling */
	void (*composite_proppatch_func)(PropMapping *prop_mapping, EContact *new_contact, EContact *cur_contact, E2kProperties *props);
	void (*composite_populate_func)(EContactField e_book_field, EContact *new_contact, void *data);
};

#define DATE_FIELD(x,z) { E_CONTACT_##x, E2K_PR_CONTACTS_##x, z, FLAG_COMPOSITE | FLAG_UNLIKEABLE, proppatch_date, populate_date }
#define EMAIL_FIELD(x,z) { E_CONTACT_##x, E2K_PR_MAPI_##x##_ADDRESS, z, FLAG_EMAIL | FLAG_COMPOSITE, proppatch_email }
#define ADDRESS_FIELD(x,z) { E_CONTACT_##x, E2K_PR_CONTACTS_##x, z, FLAG_COMPOSITE, proppatch_address, populate_address }

#define CONTACT_FIELD(x,z) { E_CONTACT_##x, E2K_PR_CONTACTS_##x, z, 0 }
#define MAPI_ID_FIELD(x,y,z) { E_CONTACT_##x, y, z, FLAG_UNLIKEABLE, NULL }
#define CAL_FIELD(x,z) { E_CONTACT_##x, E2K_PR_CALENDAR_##x, z, FLAG_UNLIKEABLE, NULL }
#define HTTPMAIL_FIELD(x,y,z) { E_CONTACT_##x, E2K_PR_HTTPMAIL_##y, z, FLAG_UNLIKEABLE | FLAG_PUT, NULL }
#define EXCHANGE_FIELD(x,y,z,w,v) { E_CONTACT_##x, E2K_PR_EXCHANGE_##y, z, FLAG_COMPOSITE, w, v }


static gboolean
value_changed (const char *old, const char *new)
{
	if (old == NULL)
		return new != NULL;
	else if (new == NULL)
		return old != NULL;
	else
		return strcmp (old, new) != 0;
}

static const EContactField email_fields[3] = {
	E_CONTACT_EMAIL_1,
	E_CONTACT_EMAIL_2,
	E_CONTACT_EMAIL_3
};

static const char *email1_props[] = {
	E2K_PR_MAPI_EMAIL_1_ENTRYID,
	E2K_PR_MAPI_EMAIL_1_ADDRTYPE,
	E2K_PR_MAPI_EMAIL_1_ADDRESS,
	E2K_PR_MAPI_EMAIL_1_DISPLAY_NAME,
	E2K_PR_CONTACTS_EMAIL1,
};
static const char *email2_props[] = {
	E2K_PR_MAPI_EMAIL_2_ENTRYID,
	E2K_PR_MAPI_EMAIL_2_ADDRTYPE,
	E2K_PR_MAPI_EMAIL_2_ADDRESS,
	E2K_PR_MAPI_EMAIL_2_DISPLAY_NAME,
	E2K_PR_CONTACTS_EMAIL2,
};
static const char *email3_props[] = {
	E2K_PR_MAPI_EMAIL_3_ENTRYID,
	E2K_PR_MAPI_EMAIL_3_ADDRTYPE,
	E2K_PR_MAPI_EMAIL_3_ADDRESS,
	E2K_PR_MAPI_EMAIL_3_DISPLAY_NAME,
	E2K_PR_CONTACTS_EMAIL3,
};

static const char **email_props[] = {
	email1_props, email2_props, email3_props
};

static void
proppatch_email (PropMapping *prop_mapping,
		 EContact *new_contact, EContact *cur_contact,
		 E2kProperties *props)
{
	gboolean changed;
	char *new_email, *cur_email;
	int field, prop, emaillisttype = 0;

	/* We do all three email addresses (plus some additional data)
	 * when invoked for E_CONTACT_EMAIL_1. So skip EMAIL_2
	 * and EMAIL_3.
	 */
	if (prop_mapping->field != E_CONTACT_EMAIL_1)
		return;

	for (field = 0; field < 3; field++) {
		new_email = e_contact_get (new_contact, email_fields[field]);
		cur_email = cur_contact ? e_contact_get (cur_contact, email_fields[field]) : NULL;

		if (new_email)
			emaillisttype |= (1 << field);

		changed = value_changed (cur_email, new_email);
		g_free (cur_email);

		if (!changed) {
			g_free (new_email);
			continue;
		}

		if (!new_email || !*new_email) {
			g_free (new_email);
			if (!cur_contact)
				continue;

			for (prop = 0; prop < 5; prop++) {
				e2k_properties_remove (
					props, email_props[field][prop]);
			}
			continue;
		}

		/* Clear originalentryid */
		e2k_properties_remove (props, email_props[field][0]);

		/* type is SMTP */
		e2k_properties_set_string (props, email_props[field][1],
					   g_strdup ("SMTP"));

		for (prop = 2; prop < 5; prop++) {
			e2k_properties_set_string (props,
						   email_props[field][prop],
						   g_strdup (new_email));
		}
		g_free (new_email);
	}

	e2k_properties_set_int (props, E2K_PR_MAPI_EMAIL_LIST_TYPE,
				emaillisttype);

	if (emaillisttype) {
		GPtrArray *fields;

		fields = g_ptr_array_new ();
		for (field = 0; field < 3; field++) {
			if (emaillisttype & (1 << field))
				g_ptr_array_add (fields, g_strdup_printf ("%d", field));
		}
		e2k_properties_set_int_array (props,
					      E2K_PR_MAPI_EMAIL_ADDRESS_LIST,
					      fields);
	} else if (cur_contact)
		e2k_properties_remove (props, E2K_PR_MAPI_EMAIL_ADDRESS_LIST);
}

static void
proppatch_date (PropMapping *prop_mapping,
		EContact *new_contact, EContact *cur_contact,
		E2kProperties *props)
{
	gboolean changed;
	EContactDate *new_dt, *cur_dt;
	time_t tt;
	struct tm then;

	new_dt = e_contact_get (new_contact, prop_mapping->field);
	if (cur_contact)
		cur_dt = e_contact_get (cur_contact, prop_mapping->field);
	else
		cur_dt = NULL;

	changed = !e_contact_date_equal (cur_dt, new_dt);
	e_contact_date_free (cur_dt);

	if (!changed) {
		e_contact_date_free (new_dt);
		return;
	}

	memset (&then, 0, sizeof(then));
	then.tm_year = new_dt->year - 1900;
	then.tm_mon  = new_dt->month - 1;
	then.tm_mday = new_dt->day;
	then.tm_isdst = -1;
	tt = mktime (&then);

	e2k_properties_set_date (props, prop_mapping->prop_name,
				 e2k_make_timestamp (tt));
	e_contact_date_free (new_dt);
}

static void
populate_date(EContactField field, EContact *new_contact, void *data)
{
	char *date = (char*)data;
	time_t tt;
	struct tm *then;
	EContactDate dt;

	tt = e2k_parse_timestamp (date);

	then = gmtime (&tt);

	dt.year = then->tm_year + 1900;
	dt.month = then->tm_mon + 1;
	dt.day = then->tm_mday;
	
	e_contact_set (new_contact, field, &dt);
}

enum addressprop {
	ADDRPROP_POSTOFFICEBOX,
	ADDRPROP_STREET,
	ADDRPROP_CITY,
	ADDRPROP_STATE,
	ADDRPROP_POSTALCODE,
	ADDRPROP_COUNTRY,
	ADDRPROP_LAST
};

static char *homeaddrpropnames[] = {
	E2K_PR_CONTACTS_HOME_PO_BOX,
	E2K_PR_CONTACTS_HOME_STREET,
	E2K_PR_CONTACTS_HOME_CITY,
	E2K_PR_CONTACTS_HOME_STATE,
	E2K_PR_CONTACTS_HOME_ZIP,
	E2K_PR_CONTACTS_HOME_COUNTRY,
};

static char *otheraddrpropnames[] = {
	E2K_PR_CONTACTS_OTHER_PO_BOX,
	E2K_PR_CONTACTS_OTHER_STREET,
	E2K_PR_CONTACTS_OTHER_CITY,
	E2K_PR_CONTACTS_OTHER_STATE,
	E2K_PR_CONTACTS_OTHER_ZIP,
	E2K_PR_CONTACTS_OTHER_COUNTRY,
};

static char *workaddrpropnames[] = {
	E2K_PR_CONTACTS_WORK_PO_BOX,
	E2K_PR_CONTACTS_WORK_STREET,
	E2K_PR_CONTACTS_WORK_CITY,
	E2K_PR_CONTACTS_WORK_STATE,
	E2K_PR_CONTACTS_WORK_ZIP,
	E2K_PR_CONTACTS_WORK_COUNTRY,
};

static void
proppatch_address (PropMapping *prop_mapping, 
		   EContact *new_contact, EContact *cur_contact,
		   E2kProperties *props)
{
	const EContactAddress *new_address, *cur_address = NULL;
	char *new_addrprops[ADDRPROP_LAST], *cur_addrprops[ADDRPROP_LAST];
	char **propnames, *value;
	int i;

	new_address = e_contact_get (new_contact, prop_mapping->field);
	if (cur_contact)
		cur_address = e_contact_get (cur_contact, prop_mapping->field);

	switch (prop_mapping->field) {
	case E_CONTACT_ADDRESS_HOME:
		propnames = homeaddrpropnames;
		break;

	case E_CONTACT_ADDRESS_WORK:
		propnames = workaddrpropnames;
		break;

	case E_CONTACT_ADDRESS_OTHER:
	default:
		propnames = otheraddrpropnames;
		break;
	}

	if (!new_address) {
		if (cur_address) {
			for (i = 0; i < ADDRPROP_LAST; i ++)
				e2k_properties_remove (props, propnames[i]);
		}
		return;
	}

	new_addrprops [ADDRPROP_POSTOFFICEBOX]		= new_address->po;
	new_addrprops [ADDRPROP_STREET]			= new_address->street;
	new_addrprops [ADDRPROP_CITY]			= new_address->locality;
	new_addrprops [ADDRPROP_STATE]			= new_address->region;
	new_addrprops [ADDRPROP_POSTALCODE]		= new_address->code;
	new_addrprops [ADDRPROP_COUNTRY]		= new_address->country;
	if (cur_address) {
		cur_addrprops [ADDRPROP_POSTOFFICEBOX]	= cur_address->po;
		cur_addrprops [ADDRPROP_STREET]		= cur_address->street;
		cur_addrprops [ADDRPROP_CITY]		= cur_address->locality;
		cur_addrprops [ADDRPROP_STATE]		= cur_address->region;
		cur_addrprops [ADDRPROP_POSTALCODE]	= cur_address->code;
		cur_addrprops [ADDRPROP_COUNTRY]	= cur_address->country;
	}

	for (i = 0; i < ADDRPROP_LAST; i ++) {
		if (!new_addrprops[i]) {
			if (cur_addrprops[i])
				e2k_properties_remove (props, propnames[i]);
			continue;
		}

		if (cur_address && cur_addrprops[i] &&
		    !strcmp (new_addrprops[i], cur_addrprops[i])) {
			/* they're the same */
			continue;
		}

		if (i == ADDRPROP_STREET && new_address->ext) {
			value = g_strdup_printf ("%s%s", new_addrprops[i],
						 new_address->ext);
		} else
			value = g_strdup (new_addrprops[i]);
		e2k_properties_set_string (props, propnames[i], value);
	}
}

static void
populate_address (EContactField field, EContact *new_contact, void *data)
{
	EAddressWestern *waddr;
	EContactAddress addr;

	waddr = e_address_western_parse ((const char *)data);
	addr.address_format = "us"; /* FIXME? */
	addr.po = waddr->po_box;
	addr.ext = waddr->extended;
	addr.street = waddr->street;
	addr.locality = waddr->locality;
	addr.region = waddr->region;
	addr.code = waddr->postal_code;
	addr.country = waddr->country;

	e_contact_set (new_contact, field, &addr);
	e_address_western_free (waddr);
}

static PropMapping
prop_mappings[] = {
	CONTACT_FIELD (FULL_NAME, "full_name"),
	CONTACT_FIELD (FAMILY_NAME, "family_name"),
	CONTACT_FIELD (GIVEN_NAME, "given_name"),
/*	CONTACT_FIELD (ADDITIONAL_NAME, "middle_name"), FIXME */
/*	CONTACT_FIELD (NAME_SUFFIX, "name_suffix"), FIXME */
	CONTACT_FIELD (TITLE, "title"),
	CONTACT_FIELD (ORG, "org"),
	CONTACT_FIELD (FILE_AS, "file_as"),

	CONTACT_FIELD (PHONE_CALLBACK, "callback_phone"),
	CONTACT_FIELD (PHONE_BUSINESS_FAX, "business_fax"),
	CONTACT_FIELD (PHONE_HOME_FAX, "home_fax"),
	CONTACT_FIELD (PHONE_HOME, "home_phone"),
	CONTACT_FIELD (PHONE_HOME_2, "home_phone_2"),
	CONTACT_FIELD (PHONE_ISDN, "isdn"),
	CONTACT_FIELD (PHONE_MOBILE, "mobile_phone"),
	CONTACT_FIELD (PHONE_COMPANY, "company_phone"),
	CONTACT_FIELD (PHONE_OTHER_FAX, "other_fax"),
	CONTACT_FIELD (PHONE_PAGER, "pager"),
	CONTACT_FIELD (PHONE_BUSINESS, "business_phone"),
	CONTACT_FIELD (PHONE_BUSINESS_2, "business_phone_2"),
	CONTACT_FIELD (PHONE_TELEX, "telex"),
	CONTACT_FIELD (PHONE_TTYTDD, "tty"),
	CONTACT_FIELD (PHONE_ASSISTANT, "assistant_phone"),
	CONTACT_FIELD (PHONE_CAR, "car_phone"),
	CONTACT_FIELD (PHONE_OTHER, "other_phone"),
	MAPI_ID_FIELD (PHONE_RADIO, PR_RADIO_TELEPHONE_NUMBER, "radio"),
	MAPI_ID_FIELD (PHONE_PRIMARY, PR_PRIMARY_TELEPHONE_NUMBER, "primary_phone"),

	EMAIL_FIELD (EMAIL_1, "email_1"),
	EMAIL_FIELD (EMAIL_2, "email_2"),
	EMAIL_FIELD (EMAIL_3, "email_3"),

	ADDRESS_FIELD (ADDRESS_WORK, "business_address"),
	ADDRESS_FIELD (ADDRESS_HOME, "home_address"),
	ADDRESS_FIELD (ADDRESS_OTHER, "other_address"),

	CONTACT_FIELD (HOMEPAGE_URL, "url"),
	CONTACT_FIELD (ORG_UNIT, "org_unit"),
	CONTACT_FIELD (OFFICE, "office"),
	CONTACT_FIELD (ROLE, "role"),
	CONTACT_FIELD (MANAGER, "manager"),
	CONTACT_FIELD (ASSISTANT, "assistant"),
	CONTACT_FIELD (NICKNAME, "nickname"),
	CONTACT_FIELD (SPOUSE, "spouse"),

	DATE_FIELD (BIRTH_DATE, "birth_date"),
	DATE_FIELD (ANNIVERSARY, "anniversary"),
	CAL_FIELD (FREEBUSY_URL, "fburl"),

	HTTPMAIL_FIELD (NOTE, TEXT_DESCRIPTION, "note"),

#ifdef ENABLE_CATEGORIES
	/* this doesn't work at the moment */
	EXCHANGE_FIELD (CATEGORIES, KEYWORDS, "categories", proppatch_categories, populate_categories)
#endif
};
static int num_prop_mappings = sizeof(prop_mappings) / sizeof (prop_mappings[0]);

static const char *
contact_name (EContact *contact)
{
	const char *contact_name;

	contact_name = e_contact_get_const (contact, E_CONTACT_FULL_NAME);
	if (contact_name && *contact_name)
		return contact_name;
	contact_name = e_contact_get_const (contact, E_CONTACT_FILE_AS);
	if (contact_name && *contact_name)
		return contact_name;
	contact_name = e_contact_get_const (contact, E_CONTACT_EMAIL_1);
	if (contact_name && *contact_name)
		return contact_name;
	contact_name = e_contact_get_const (contact, E_CONTACT_EMAIL_2);
	if (contact_name && *contact_name)
		return contact_name;
	contact_name = e_contact_get_const (contact, E_CONTACT_EMAIL_3);
	if (contact_name && *contact_name)
		return contact_name;
	contact_name = e_contact_get_const (contact, E_CONTACT_ORG);
	if (contact_name && *contact_name)
		return contact_name;
	contact_name = e_contact_get_const (contact, E_CONTACT_TITLE);
	if (contact_name && *contact_name)
		return contact_name;

	return NULL;
}

static E2kProperties *
props_from_contact (EBookBackendOGo *be,
		    EContact *contact,
		    EContact *old_contact)
{
	E2kProperties *props;
	int i;

	props = e2k_properties_new ();

	if (!old_contact) {
		const char *subject;

		subject = contact_name (contact);

		/* Set up some additional fields when creating a new contact */
		e2k_properties_set_string (
			props, E2K_PR_EXCHANGE_MESSAGE_CLASS,
			g_strdup ("IPM.Contact"));
		e2k_properties_set_string (
			props, E2K_PR_HTTPMAIL_SUBJECT,
			g_strdup (subject ? subject : ""));

		e2k_properties_set_int (props, E2K_PR_MAPI_SIDE_EFFECTS, 16);
		e2k_properties_set_int (props, PR_ACTION, 512);
		e2k_properties_set_bool (props, E2K_PR_OUTLOOK_CONTACT_JOURNAL, FALSE);
		e2k_properties_set_bool (props, E2K_PR_MAPI_SENSITIVITY, 0);
	}
	
	for (i = 0; i < num_prop_mappings; i ++) {
		/* handle composite attributes here (like addresses) */
		if (prop_mappings[i].flags & FLAG_COMPOSITE) {
			prop_mappings[i].composite_proppatch_func (&prop_mappings[i], contact, old_contact, props);
		} else if (prop_mappings[i].flags & FLAG_PUT) {
			continue; /* FIXME */
		} else {
			const char *new_value, *current_value;

			new_value = e_contact_get_const (contact, prop_mappings[i].field);
			if (new_value && !*new_value)
				new_value = NULL;
			current_value = old_contact ? e_contact_get_const (old_contact, prop_mappings[i].field) : NULL;
			if (current_value && !*current_value)
				current_value = NULL;

			if (value_changed (current_value, new_value)) {
				if (new_value) {
					e2k_properties_set_string (
						props,
						prop_mappings[i].prop_name,
						g_strdup (new_value));
				} else {
					e2k_properties_remove (
						props,
						prop_mappings[i].prop_name);
				}
			}
		}
	}

	if (e2k_properties_empty (props)) {
		e2k_properties_free (props);
		return NULL;
	}

	return props;
}

static void
get_email_field_from_props (PropMapping *prop_mapping, E2kProperties *props,
			    EContact *contact, char *data)
{
	char *emailtype;
	char *typeselector;

	/* here's where we do the super involved
	   conversion from local email addresses to
	   internet addresses for display in
	   evolution. */
	if (prop_mapping->field == E_CONTACT_EMAIL_1)
		typeselector = E2K_PR_MAPI_EMAIL_1_ADDRTYPE;
	else if (prop_mapping->field == E_CONTACT_EMAIL_2)
		typeselector = E2K_PR_MAPI_EMAIL_2_ADDRTYPE;
	else if (prop_mapping->field == E_CONTACT_EMAIL_3)
		typeselector = E2K_PR_MAPI_EMAIL_3_ADDRTYPE;
	else {
		g_warning ("invalid email field");
		return;
	}

	emailtype = e2k_properties_get_prop (props, typeselector);
	if (!emailtype) {
		g_warning ("no email info for \"%s\"", data);
		return;
	}
	if (!strcmp (emailtype, "SMTP")) {
		/* it's a normal email address, just set the value as we usually would */
		e_contact_set (contact, prop_mapping->field, data);
	}
	else if (emailtype[0] == '\0') {
		/* empty emailtype == no email address here */
		return;
	}
	else {
		g_warning ("email address type `%s' not handled, using the value as is", emailtype);
		e_contact_set (contact, prop_mapping->field, data);
	}

	return;
}

/* Exchange uses \r in some strings and \r\n in others. EContact always
 * wants just \n.
 */
static char *
unixify (const char *string)
{
	char *out = g_malloc (strlen (string) + 1), *d = out;
	const char *s = string;

	do {
		if (*s == '\r') {
			if (*(s + 1) != '\n')
				*d++ = '\n';
		} else
			*d++ = *s;
	} while (*s++);

	return out;
}

static EContact *
e_contact_from_props (const gchar *href, E2kProperties *props)
{
	EContact *contact;
	char *data, *body;
	const char *filename;
	int i, len;

	contact = e_contact_new ();

	e_contact_set (contact, E_CONTACT_UID, (gchar *)href);
	for (i = 0; i < num_prop_mappings; i ++) {
		data = e2k_properties_get_prop (props,
						prop_mappings[i].prop_name);
		if (!data)
			continue;

		if (prop_mappings[i].flags & FLAG_EMAIL) {
			get_email_field_from_props (&prop_mappings[i],props,
						    contact, data);
		} else if (prop_mappings[i].flags & FLAG_COMPOSITE) {
			prop_mappings[i].composite_populate_func (
				prop_mappings[i].field, contact, data);
		} else {
			char *unix_data;

			unix_data = strchr (data, '\r') ?
				unixify (data) : data;
			e_contact_set (contact, prop_mappings[i].field,
				       unix_data);
			if (unix_data != data)
				g_free (unix_data);
		}
	}

	return contact;
}


static void
e_book_backend_ogo_modify_contact (EBookBackend *backend,
				   EDataBook *book,
				   guint32 opid,
				   const char *vcard)
{
	EBookBackendOGo *ogo = E_BOOK_BACKEND_OGO (backend);
	EContact *modified_contact, *cached_contact;
	gchar *href;
	gchar *uid;
	OGoConnectionStatus status;
	E2kProperties *props;
	gint nprops;
	gchar *e2k_href;
	
	if (!ogo->priv->conn) {
		e_data_book_respond_modify (book, opid, GNOME_Evolution_Addressbook_AuthenticationRequired, NULL);
		return;
	}

	modified_contact = e_contact_new_from_vcard (vcard);

	uid = e_contact_get_const (modified_contact, E_CONTACT_UID);

	cached_contact = ogo_book_backend_cache_get_contact (ogo->priv->cache, uid);

	if (!cached_contact) {
		e_data_book_respond_modify (book, opid, GNOME_Evolution_Addressbook_ContactNotFound, NULL);
		g_object_unref (modified_contact);
		return;
	}

	get_ogo_info (cached_contact, &href, NULL);
	e2k_href = get_e2k_uri (href);
	g_free (href);

	props = props_from_contact (ogo, modified_contact, NULL); 
	
	status = ogo_connection_proppatch (ogo->priv->conn, e2k_href,
					   props, FALSE, NULL);
	g_free (e2k_href);
	
	if (status == OGO_CONNECTION_STATUS_OK) {
		/* FIXME: Need to add new info about the contact here */
		ogo_book_backend_cache_put_contact (ogo->priv->cache, modified_contact);
		e_data_book_respond_modify (book, opid, GNOME_Evolution_Addressbook_Success, modified_contact);
	} else {
		e_data_book_respond_modify (book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
	}

	g_object_unref (modified_contact);
}

static void
e_book_backend_ogo_get_contact (EBookBackend *backend,
				EDataBook *book,
				guint32 opid,
				const char *id)
{
	EBookBackendOGo *bf;
	int              db_error = 0;

	g_warning ("%s: implement me!", __PRETTY_FUNCTION__);
}

static void
e_book_backend_ogo_get_contact_list (EBookBackend *backend,
				     EDataBook        *book,
				     guint32           opid,
				     const char       *query)
{
	EBookBackendOGo     *ogo = E_BOOK_BACKEND_OGO (backend);
	EBookBackendSExp    *card_sexp = NULL;
	gboolean             search_needed;
	const char          *search = query;
	GList               *vcard_list = NULL;
	gboolean             match_needed;
	GList               *contacts, *list;
	OGoConnectionStatus  status;

	g_message ("Looking for contacts ...\n");


	if (!ogo->priv->conn) {
		e_data_book_respond_get_contact_list (book, opid, GNOME_Evolution_Addressbook_AuthenticationRequired, NULL);
		return;
	}
	
	g_print ("query is: %s\n", query);

	card_sexp = e_book_backend_sexp_new (query);
	if (!card_sexp) {
		e_data_book_respond_get_contact_list (book, opid, GNOME_Evolution_Addressbook_InvalidQuery,
						      vcard_list);
		return;
	}

	if (strcmp (query, "(contains \"x-evolution-any-field\" \"\")") != 0)
		match_needed = TRUE;
	else
		match_needed = FALSE;
		

	contacts = ogo_book_backend_cache_get_contacts (ogo->priv->cache);
	for (list = contacts; list != NULL; list = list->next) {
		EContact *contact = list->data;

		if (match_needed) {
			if (e_book_backend_sexp_match_contact (card_sexp, contact))
				vcard_list = g_list_append (vcard_list, e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30));
			g_object_unref (contact);
		}
		else 
			vcard_list = g_list_append (vcard_list, e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30));
		
	}
	
	g_list_free (contacts);
	e_data_book_respond_get_contact_list (book, opid, GNOME_Evolution_Addressbook_Success,
					      vcard_list);

	g_object_unref (card_sexp);
}

typedef struct {
	EBookBackendOGo *ogo;
	GMutex          *mutex;
	GCond           *cond;
	GThread         *thread;
	gboolean         stopped;
} OGoBackendSearchClosure;

static void
closure_destroy (OGoBackendSearchClosure *closure)
{
	d(printf ("destroying search closure\n"));
	g_mutex_free (closure->mutex);
	g_cond_free (closure->cond);
	g_free (closure);
}

static OGoBackendSearchClosure*
init_closure (EDataBookView *book_view, EBookBackendOGo *ogo)
{
	OGoBackendSearchClosure *closure = g_new (OGoBackendSearchClosure, 1);

	closure->ogo = ogo;
	closure->mutex = g_mutex_new ();
	closure->cond = g_cond_new ();
	closure->thread = NULL;
	closure->stopped = FALSE;

	g_object_set_data_full (G_OBJECT (book_view), "EBookBackendOGo.BookView::closure",
				closure, (GDestroyNotify)closure_destroy);

	return closure;
}

static OGoBackendSearchClosure*
get_closure (EDataBookView *book_view)
{
	return g_object_get_data (G_OBJECT (book_view), "EBookBackendOGo.BookView::closure");
}

static gpointer
book_view_thread (gpointer data)
{
	EDataBookView           *book_view = data;
	OGoBackendSearchClosure *closure = get_closure (book_view);
	EBookBackendOGo         *ogo = closure->ogo;
	gboolean                 stopped = FALSE;
	GList                   *list, *tmp;

	bonobo_object_ref (book_view);
	g_mutex_lock (closure->mutex);
	g_cond_signal (closure->cond);
	g_mutex_unlock (closure->mutex);

	if (!ogo->priv->conn) {
		e_data_book_view_notify_complete (book_view, GNOME_Evolution_Addressbook_AuthenticationRequired);
		bonobo_object_unref (book_view);
		return NULL;
	}
	
	e_data_book_view_notify_status_message (book_view, _("Searching..."));

	list = ogo_book_backend_cache_get_contacts (ogo->priv->cache);

	for (tmp = list; tmp != NULL; tmp = tmp->next) {
		EContact *contact = tmp->data;
		
		g_mutex_lock (closure->mutex);
		stopped = closure->stopped;
		g_mutex_unlock (closure->mutex);

		if (stopped)
			break;


		e_data_book_view_notify_update (book_view, contact);

		g_object_unref (contact);
	}

	g_list_free (list);
	
	if (!stopped)
		e_data_book_view_notify_complete (book_view, GNOME_Evolution_Addressbook_Success);
	
	/* unref  */
	bonobo_object_unref (book_view);

	return NULL;
}

static void
e_book_backend_ogo_start_book_view (EBookBackend  *backend,
				    EDataBookView *book_view)
{
	OGoBackendSearchClosure *closure = init_closure 
		(book_view, E_BOOK_BACKEND_OGO (backend));
	
	g_mutex_lock (closure->mutex);

	printf ("Starting contacts book view thread\n");
	closure->thread = g_thread_create (book_view_thread, book_view, FALSE, NULL);

	g_cond_wait (closure->cond, closure->mutex);

	/* at this point we know the book view thread is actually running */
	g_mutex_unlock (closure->mutex);

	e_book_backend_set_is_loaded (E_BOOK_BACKEND (backend), TRUE);
}

static void
e_book_backend_ogo_stop_book_view (EBookBackend  *backend,
				    EDataBookView *book_view)
{
	OGoBackendSearchClosure *closure = get_closure (book_view);

	g_mutex_lock (closure->mutex);
	if (!closure->stopped)
		closure->stopped = TRUE;
	g_mutex_unlock (closure->mutex);
}


static void
e_book_backend_ogo_get_changes (EBookBackend *backend,
				EDataBook       *book,
				guint32           opid,
				const char       *change_id)
{
	/* FIXME: Find the changed ids and put them in changes_out */
}

static char *
e_book_backend_ogo_extract_path_from_uri (const char *uri)
{
	g_assert (strncasecmp (uri, "ogo://", 6) == 0);

	return g_strdup (uri + 6);
}

static gboolean populate_cache (EBookBackendOGo *ogo);
static gboolean update_cache (EBookBackendOGo *ogo);

static void
e_book_backend_ogo_authenticate_user (EBookBackend *backend,
				      EDataBook    *book,
				      guint32       opid,
				      const char   *user,
				      const char   *passwd,
				      const char   *auth_method)
{
	EBookBackendOGo *ogo;
	EBookBackendOGoPrivate *priv;

	ogo = E_BOOK_BACKEND_OGO (backend);
	priv = ogo->priv;

	g_message ("Auth user now ...!!!");
	
	priv->conn = ogo_connection_new (priv->uri, user, passwd);
	
#if 0
	if(!ogo_book_backend_cache_get_marker (priv->cache))
		g_idle_add ((GSourceFunc) populate_cache, ogo);  
	else 
		g_idle_add ((GSourceFunc) update_cache, ogo);
#else
	if(!ogo_book_backend_cache_get_marker (priv->cache))
		populate_cache (ogo);
	else
		update_cache (ogo);
#endif
	
	e_data_book_respond_authenticate_user (book, opid, GNOME_Evolution_Addressbook_Success); 
}

static void
e_book_backend_ogo_get_supported_fields (EBookBackend *backend,
					 EDataBook        *book,
					 guint32           opid)
{
	GList *fields = NULL;
	int i;

	for (i = 0; i < num_prop_mappings; i ++) {
		if (prop_mappings[i].e_book_field) {
			fields = g_list_prepend (fields,
						 g_strdup (e_contact_field_name(prop_mappings[i].field)));
		}
	}
	
	e_data_book_respond_get_supported_fields (book, opid,
						  GNOME_Evolution_Addressbook_Success,
						  fields);
}

static void
e_book_backend_ogo_get_required_fields (EBookBackend *backend,
					EDataBook    *book,
					guint32       opid)
{
	GList *fields = NULL;
	int i;

	/* FIXME: Need to figure out what fields are supported here */
	fields = g_list_append (fields , g_strdup (e_contact_field_name (E_CONTACT_FILE_AS)));
	fields = g_list_append (fields , g_strdup (e_contact_field_name (E_CONTACT_FULL_NAME)));
	fields = g_list_append (fields , g_strdup (e_contact_field_name (E_CONTACT_GIVEN_NAME)));
	fields = g_list_append (fields , g_strdup (e_contact_field_name (E_CONTACT_FAMILY_NAME)));
	
	e_data_book_respond_get_supported_fields (book, opid,
						  GNOME_Evolution_Addressbook_Success,
						  fields);
	g_list_free (fields);
}

static gchar *
get_uid_cb (OGoItem *item)
{
	EContact *contact;
	gchar *uid;
	
	contact = e_contact_new_from_vcard (item->contents);

	if (!contact)
		return NULL;

	uid = e_contact_get_const (contact, E_CONTACT_UID);

	if (!uid) {
		g_object_unref (contact);
		return NULL;
	} else {
		uid = g_strdup (uid);
		
		g_object_unref (contact);

		return uid;
	}
}

static void
add_ogo_props (EContact *contact,
	       const gchar *etag,
	       const gchar *href)
{
	EVCard *card;
	EVCardAttribute *attr;
	
	card = E_VCARD (contact);

	attr = e_vcard_attribute_new (NULL, OGO_ETAG_PROPERTY);
	e_vcard_attribute_add_value (attr, etag);
	e_vcard_add_attribute (card, attr);

	attr = e_vcard_attribute_new (NULL, OGO_HREF_PROPERTY);
	e_vcard_attribute_add_value (attr, href);
	e_vcard_add_attribute (card, attr);
}

static gboolean
populate_cache (EBookBackendOGo *ogo)
{
	OGoConnectionStatus status;
	GList *items, *list;
	
	status = ogo_connection_fetch_items (ogo->priv->conn, ogo->priv->item_path,
					     &items);
	
	if (status != OGO_CONNECTION_STATUS_OK) 
		return FALSE;

	for (list = items; list != NULL; list = list->next) {
		OGoItem *item = list->data;
		EContact *contact;

		contact = e_contact_new_from_vcard (item->contents);
		add_ogo_props (contact, item->etag, item->uri);

		ogo_book_backend_cache_put_contact (ogo->priv->cache, contact);
		
		g_object_unref (contact);
	}
	
	ogo_item_list_free (items);

	ogo_book_backend_cache_set_marker (ogo->priv->cache);
	
	return FALSE;
}

/* Creates a hash table so we can lookup etags from the URI quickly */
typedef struct {
	gchar *uri;
	gchar *etag;
	gboolean touched;
} HashItem;

static void
hash_item_free (HashItem *item)
{
	g_free (item->uri);
	g_free (item->etag);
	g_free (item);
}

static GHashTable *
create_item_hash (GList *items)
{
	GHashTable *table;

	table = g_hash_table_new_full (g_str_hash, g_str_equal,
				       NULL, (GDestroyNotify)hash_item_free);

	for (; items != NULL; items = items->next) {
		OGoItem *item = items->data;
		HashItem *hash_item;

		hash_item = g_new (HashItem, 1);
		hash_item->uri = g_strdup (item->uri);
		hash_item->etag = g_strdup (item->etag);
		hash_item->touched = FALSE;
		
		g_hash_table_insert (table, hash_item->uri, hash_item);
	}
	
	return table;
}

static gchar *
get_item_etag_and_href (GHashTable *table, const gchar *uri, gchar **href)
{
	HashItem *item;

	item = g_hash_table_lookup (table, uri);

	if (!item)
		return NULL;

	if (href)
		*href = item->uri;
	
	item->touched = TRUE;

	
	return item->etag;
}

static void
get_new_uris_foreach (gpointer key, gpointer value, gpointer user_data)
{
	GList **list = user_data;
	HashItem *item = value;

	if (!item->touched)
		*list = g_list_prepend (*list, item->uri);
}

static GList *
get_new_uris (GHashTable *table)
{
	GList *new_uris = NULL;

	g_hash_table_foreach (table,
			      get_new_uris_foreach, &new_uris);

	return new_uris;
}

static void
get_ogo_info (EContact *contact, gchar **href, gchar **etag)
{
	GList *attrs;
	gint n_found = 0;
	
	for (attrs = e_vcard_get_attributes (E_VCARD (contact));
	     attrs != NULL; attrs = attrs->next) {
		EVCardAttribute *attr = attrs->data;

 		if (strcmp (e_vcard_attribute_get_name (attr), OGO_ETAG_PROPERTY) == 0) {
			if (etag)
				*etag = g_strdup (e_vcard_attribute_get_value (attr));
			n_found ++;
		} else if (strcmp (e_vcard_attribute_get_name (attr), OGO_HREF_PROPERTY) == 0) {
			if (href)
				*href = g_strdup (e_vcard_attribute_get_value (attr));
			n_found ++;
		}
	}

	g_assert (n_found == 2);
}

EContact *
fetch_contact (EBookBackendOGo *ogo,
	       const gchar *href,
	       const gchar *etag)
{
	EContact *contact;
	gchar *contents;
	EUri *uri;
	
	/* FIXME: We assume here that the item hrefs point to the same server which might not be true */
	uri = e_uri_new (href);
	ogo_connection_fetch_item_absolute (ogo->priv->conn, uri->path, &contents);
	e_uri_free (uri);

	if (!contents)
		return;
	
	contact = e_contact_new_from_vcard (contents);

	g_free (contents);
	
	if (!contact)
		return NULL;

	add_ogo_props (contact, etag, href);
	
	return contact;
}

static gboolean
update_cache (EBookBackendOGo *ogo)
{
	OGoConnectionStatus status;
	GList *items, *list;
	GHashTable *item_hash;
	GList *cached_items;
	GList *updated_items = NULL;
	GList *new_items;

	status = ogo_connection_fetch_items (ogo->priv->conn, ogo->priv->item_path,
					     &items);
	
	if (status != OGO_CONNECTION_STATUS_OK) 
		return TRUE;

	item_hash = create_item_hash (items);
	ogo_item_list_free (items);

	cached_items = ogo_book_backend_cache_get_contacts (ogo->priv->cache);
	for (list = cached_items; list != NULL; list = list->next) {
		gchar *href, *etag;
		gchar *real_href, *cached_etag;
		
		EContact *contact = E_CONTACT (list->data);

		get_ogo_info (contact, &href, &etag);

		cached_etag = get_item_etag_and_href (item_hash, href, &real_href);

		if (!cached_etag) {
			const gchar *uid = e_contact_get_const (contact, E_CONTACT_UID);
			
			ogo_book_backend_cache_remove_contact (ogo->priv->cache, uid);
			e_book_backend_notify_remove (E_BOOK_BACKEND (ogo), uid);
			
		} else if (strcmp (etag, cached_etag) != 0) {
			updated_items = g_list_prepend (updated_items, real_href);
		}

		g_object_unref (contact);
	}
	g_list_free (cached_items);

	new_items = get_new_uris (item_hash);

	updated_items = g_list_concat (updated_items, new_items);
	
	/* Fetch updated items */
	for (list = updated_items; list != NULL; list = list->next) {
		gchar *href = list->data;
		EContact *modified_contact;

		modified_contact = fetch_contact (ogo, href,
						  get_item_etag_and_href (item_hash, href, NULL));
		ogo_book_backend_cache_put_contact (ogo->priv->cache, modified_contact);
		
		e_book_backend_notify_update (E_BOOK_BACKEND (ogo), modified_contact);
		g_object_unref (modified_contact);
	}

	g_list_free (updated_items);
	g_hash_table_destroy (item_hash);
	
	return FALSE;
}

static GNOME_Evolution_Addressbook_CallStatus
e_book_backend_ogo_load_source (EBookBackend *backend,
				ESource      *source,
				gboolean      only_if_exists)
{
	EBookBackendOGo *ogo;
	EBookBackendOGoPrivate *priv;
        char *uri, *item_path, *cache_uri;
	gchar **tokens;
   	const char *port;
	EUri *parsed_uri;
	int i;
	const char *use_ssl;


	e_book_backend_notify_auth_required (backend);

	ogo = E_BOOK_BACKEND_OGO (backend);
	priv = ogo->priv;
	uri =  e_source_get_uri (source);
 	if(uri == NULL)
		return  GNOME_Evolution_Addressbook_OtherError;

	tokens = g_strsplit (uri, ";", 2);
	g_free (uri);
	if (tokens[0]) 
		uri = g_strdup(tokens[0]);
	item_path = g_strdup (tokens[1]);
	if(item_path == NULL)
		return  GNOME_Evolution_Addressbook_OtherError;
	g_strfreev (tokens);

	if(uri == NULL)
		return  GNOME_Evolution_Addressbook_OtherError;

	parsed_uri = e_uri_new (uri);
	port = e_source_get_property (source, "port");

	if (port == NULL)
		port = "80";
	use_ssl = e_source_get_property (source, "use_ssl");
	if (use_ssl) 
		priv->uri = g_strconcat ("https://", parsed_uri->host,":", port, NULL );
	else 
		priv->uri = g_strconcat ("http://", parsed_uri->host,":", port, NULL );
	priv->use_ssl = g_strdup (use_ssl);
	priv->only_if_exists = only_if_exists;
	priv->item_path = item_path;
	e_uri_free (parsed_uri);
	
	e_book_backend_set_is_loaded (E_BOOK_BACKEND (backend), TRUE);
	e_book_backend_set_is_writable (E_BOOK_BACKEND(backend), TRUE);  

	cache_uri = g_strconcat (uri, item_path, NULL);
	
	g_free (uri);
	priv->cache = ogo_book_backend_cache_new (cache_uri);
	g_free (cache_uri);
	
	return GNOME_Evolution_Addressbook_Success;
}

static void
e_book_backend_ogo_remove (EBookBackend *backend,
			   EDataBook        *book,
			   guint32           opid)
{
	EBookBackendOGo *bf = E_BOOK_BACKEND_OGO (backend);
	GDir            *dir;

	/* We don't support removing address books */
}

static char *
e_book_backend_ogo_get_static_capabilities (EBookBackend *backend)
{
	return g_strdup("do-initial-query,contact-lists");
}

static GNOME_Evolution_Addressbook_CallStatus
e_book_backend_ogo_cancel_operation (EBookBackend *backend, EDataBook *book)
{
	return GNOME_Evolution_Addressbook_CouldNotCancel;
}

static gboolean
e_book_backend_ogo_construct (EBookBackendOGo *backend)
{
	g_assert (backend != NULL);
	g_assert (E_IS_BOOK_BACKEND_OGO (backend));

	if (! e_book_backend_construct (E_BOOK_BACKEND (backend)))
		return FALSE;

	return TRUE;
}

/* Connect/Disconnect mode not supported yet!  */
static void 
e_book_backend_ogo_set_mode (EBookBackend *backend, 
			     int           mode)
{
	EBookBackendOGo *ogo;
	
	g_message ("*** OGO: CHANGING MODE ****");
	ogo = E_BOOK_BACKEND_OGO (backend);
	ogo->priv->mode = mode;
	/* set_mode is called before the backend is loaded */
	if (e_book_backend_is_loaded (backend) || 1) {
		/* Mode not supported yet */
		if (mode == GNOME_Evolution_Addressbook_MODE_LOCAL) {
			g_message ("*** OGO: LOCAL MODE NOT SUPPORTED YET ****");
			e_book_backend_notify_writable (backend, FALSE);
			e_book_backend_notify_connection_status (backend, FALSE);
			if (ogo->priv->conn) {
				g_object_unref (ogo->priv->conn);
				ogo->priv->conn=NULL;
			}
		}
		else if (mode == GNOME_Evolution_Addressbook_MODE_REMOTE) {
			g_message ("*** OGO: WE ARE IN REMOTE MODE ****");
			e_book_backend_notify_writable (backend, TRUE);
			e_book_backend_notify_connection_status (backend, TRUE);
			e_book_backend_notify_auth_required (backend);
		}
	} else {
		g_message ("*** OGO: BACKEND MUST BE LOADED ****");
	}
}

/**
 * e_book_backend_ogo_new:
 */
EBookBackend *
e_book_backend_ogo_new (void)
{
	EBookBackendOGo *backend;

	backend = g_object_new (E_TYPE_BOOK_BACKEND_OGO, NULL);

	if (! e_book_backend_ogo_construct (backend)) {
		g_object_unref (backend);

		return NULL;
	}

	return E_BOOK_BACKEND (backend);
}

static void
e_book_backend_ogo_dispose (GObject *object)
{
	EBookBackendOGo *ogo;

	ogo = E_BOOK_BACKEND_OGO (object);

	if (ogo->priv->uri) {
		g_free (ogo->priv->uri);
		ogo->priv->uri = NULL;
	}

	if (ogo->priv->item_path) {
		g_free (ogo->priv->item_path);
		ogo->priv->item_path = NULL;
	}

	if (ogo->priv->cache) {
		g_object_unref (ogo->priv->cache);
		ogo->priv->cache = NULL;
	}

	if (ogo->priv->use_ssl) {
		g_free (ogo->priv->use_ssl);
		ogo->priv->use_ssl = NULL;
	}

	G_OBJECT_CLASS (parent_class)->dispose (object);	
}

static void
e_book_backend_ogo_class_init (EBookBackendOGoClass *klass)
{
	GObjectClass    *object_class = G_OBJECT_CLASS (klass);
	EBookBackendClass *backend_class;
	int i;
	
	/* Static initialization */
	field_names_array = g_ptr_array_new ();
	g_ptr_array_add (field_names_array, E2K_PR_DAV_UID);
	g_ptr_array_add (field_names_array, E2K_PR_DAV_LAST_MODIFIED);
	g_ptr_array_add (field_names_array, E2K_PR_DAV_CREATION_DATE);
	g_ptr_array_add (field_names_array, E2K_PR_MAPI_EMAIL_1_ADDRTYPE);
	g_ptr_array_add (field_names_array, E2K_PR_MAPI_EMAIL_2_ADDRTYPE);
	g_ptr_array_add (field_names_array, E2K_PR_MAPI_EMAIL_3_ADDRTYPE);
	g_ptr_array_add (field_names_array, E2K_PR_HTTPMAIL_HAS_ATTACHMENT);
	for (i = 0; i < num_prop_mappings; i ++)
		g_ptr_array_add (field_names_array, prop_mappings[i].prop_name);
	field_names = (const char **)field_names_array->pdata;
	n_field_names = field_names_array->len;
	
	parent_class = g_type_class_peek_parent (klass);

	backend_class = E_BOOK_BACKEND_CLASS (klass);

	/* Set the virtual methods. */
	backend_class->load_source             = e_book_backend_ogo_load_source;
	backend_class->get_static_capabilities = e_book_backend_ogo_get_static_capabilities;
 
	backend_class->create_contact          = e_book_backend_ogo_create_contact;
	backend_class->remove_contacts         = e_book_backend_ogo_remove_contacts;
	backend_class->modify_contact          = e_book_backend_ogo_modify_contact;
	backend_class->get_contact             = e_book_backend_ogo_get_contact;
	backend_class->get_contact_list        = e_book_backend_ogo_get_contact_list;
	backend_class->start_book_view         = e_book_backend_ogo_start_book_view;
	backend_class->stop_book_view          = e_book_backend_ogo_stop_book_view;
	backend_class->get_changes             = e_book_backend_ogo_get_changes;
	backend_class->authenticate_user       = e_book_backend_ogo_authenticate_user;
	backend_class->get_supported_fields    = e_book_backend_ogo_get_supported_fields;
	backend_class->get_required_fields     = e_book_backend_ogo_get_required_fields;
	backend_class->cancel_operation        = e_book_backend_ogo_cancel_operation;
	backend_class->remove                  = e_book_backend_ogo_remove;
	backend_class->set_mode                 = e_book_backend_ogo_set_mode;
	
	object_class->dispose                 = e_book_backend_ogo_dispose;

	g_type_class_add_private (object_class, sizeof (EBookBackendOGoPrivate));
}

static void
e_book_backend_ogo_init (EBookBackendOGo *backend)
{
	EBookBackendOGoPrivate *priv;

	backend->priv = G_TYPE_INSTANCE_GET_PRIVATE (backend, E_TYPE_BOOK_BACKEND_OGO,
						     EBookBackendOGoPrivate);
}

/**
 * e_book_backend_ogo_get_type:
 */
GType
e_book_backend_ogo_get_type (void)
{
	static GType type = 0;

	if (! type) {
		GTypeInfo info = {
			sizeof (EBookBackendOGoClass),
			NULL, /* base_class_init */
			NULL, /* base_class_finalize */
			(GClassInitFunc)  e_book_backend_ogo_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (EBookBackendOGo),
			0,    /* n_preallocs */
			(GInstanceInitFunc) e_book_backend_ogo_init
		};

		type = g_type_register_static (E_TYPE_BOOK_BACKEND, "EBookBackendOGo", &info, 0);
	}

	return type;
}
