/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* camel-ogo-store.c : class for a ogo store */

/* 
 * Authors:
 *   Dan Winship <danw@ximian.com>
 *   Michael Zucchi <notzed@ximian.com>
 *
 * Copyright (C) 2000-2002 Ximian, Inc. (www.ximian.com)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of version 2 of the GNU General Public 
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "camel-ogo-store.h"
#include <camel/camel-session.h>
#include <utils/ogo-connection.h>

#define _(x) (x)

static CamelStoreClass *parent_class = NULL;

static void finalize (CamelObject *object);

static gboolean ogo_connect (CamelService *service, CamelException *ex);
static gboolean ogo_disconnect (CamelService *service, gboolean clean, CamelException *ex);
static GList *query_auth_types (CamelService *service, CamelException *ex);

static CamelFolder *get_folder (CamelStore *store, const char *folder_name, 
				guint32 flags, CamelException *ex);

static CamelFolder *get_trash  (CamelStore *store, CamelException *ex);

static void
camel_ogo_store_class_init (CamelOGoStoreClass *camel_ogo_store_class)
{
	CamelServiceClass *camel_service_class =
		CAMEL_SERVICE_CLASS (camel_ogo_store_class);
	CamelStoreClass *camel_store_class =
		CAMEL_STORE_CLASS (camel_ogo_store_class);

	parent_class = CAMEL_STORE_CLASS (camel_type_get_global_classfuncs (camel_store_get_type ()));
	
	/* virtual method overload */
	camel_service_class->query_auth_types = query_auth_types;
	camel_service_class->connect = ogo_connect;
	camel_service_class->disconnect = ogo_disconnect;

	camel_store_class->get_folder = get_folder;
	camel_store_class->get_trash = get_trash;
}



static void
camel_ogo_store_init (gpointer object, gpointer klass)
{
	;
}

CamelType
camel_ogo_store_get_type (void)
{
	static CamelType camel_ogo_store_type = CAMEL_INVALID_TYPE;

	if (!camel_ogo_store_type) {
		camel_ogo_store_type = camel_type_register (CAMEL_STORE_TYPE,
							     "CamelOGoStore",
							     sizeof (CamelOGoStore),
							     sizeof (CamelOGoStoreClass),
							     (CamelObjectClassInitFunc) camel_ogo_store_class_init,
							     NULL,
							     (CamelObjectInitFunc) camel_ogo_store_init,
							     finalize);
	}

	return camel_ogo_store_type;
}

static void
finalize (CamelObject *object)
{
	CamelOGoStore *ogo_store = CAMEL_OGO_STORE (object);

	/* force disconnect so we dont have it run later, after we've cleaned up some stuff */
	/* SIGH */

	camel_service_disconnect((CamelService *)ogo_store, TRUE, NULL);
}


static GList *
query_auth_types (CamelService *service, CamelException *ex)
{
	return NULL;
}


static gboolean
try_authenticate (CamelService *service, gboolean reprompt, const char *errmsg, CamelException *ex)
{
	CamelOGoStore *store = (CamelOGoStore *)service;
	gchar *uri;
	OGoConnectionStatus status;
	const gchar *auth_domain;

 	/* It might be nice to fetch the auth domain from the camel info */
	/*auth_domain = camel_url_get_param (service->url, "auth-domain");*/
	auth_domain = "OpenGroupware.org";
	
	if (!service->url->passwd) {
		char *prompt;
		guint32 flags = CAMEL_SESSION_PASSWORD_SECRET;
		
		if (reprompt)
			flags |= CAMEL_SESSION_PASSWORD_REPROMPT;
		
		prompt = g_strdup_printf (_("%sPlease enter the password for %s on host %s"),
					  errmsg ? errmsg : "",
					  service->url->user,
					  service->url->host);
		service->url->passwd = camel_session_get_password (camel_service_get_session (service), service, auth_domain,
								   prompt, "password", flags, ex);
		g_free (prompt);
		if (!service->url->passwd)
			return FALSE;
	}

	uri = g_strdup_printf ("http://%s", service->url->host);

	status = ogo_try_authenticate (uri, service->url->user, service->url->passwd);
	g_free (uri);

	switch (status) {
	case OGO_CONNECTION_STATUS_OK:
		break;
	case OGO_CONNECTION_STATUS_ACCESS_DENIED:
		camel_exception_setv (ex, CAMEL_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				      _("Unable to connect to OpenGroupware.org server %s.\n"
					"Invalid username or password"),
				      CAMEL_SERVICE (store)->url->host);
		break;
	default:
		camel_exception_setv (ex, CAMEL_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				      _("Unable to connect to OpenGroupware.org server %s.\n"
					"Unknown status error %d"), status,
				      CAMEL_SERVICE (store)->url->host);
	}

	return (status == OGO_CONNECTION_STATUS_OK);
}

static gboolean
ogo_connect (CamelService *service, CamelException *ex)
{
	CamelOGoStore *store = (CamelOGoStore *)service;
	gboolean reprompt = FALSE;
	CamelSession *session;
	char *errbuf = NULL;
	int status;
	ExceptionId id;

	session = camel_service_get_session (service);

	do {
		status = try_authenticate (service, reprompt, errbuf, ex);
		g_free (errbuf);
		errbuf = NULL;

		id = camel_exception_get_id (ex);
		/* we only re-prompt if we failed to authenticate, any other error and we just abort */
		if (id == CAMEL_EXCEPTION_SERVICE_CANT_AUTHENTICATE) {
			errbuf = g_strdup_printf ("%s\n\n", camel_exception_get_description (ex));
			g_free (service->url->passwd);
			service->url->passwd = NULL;
			reprompt = TRUE;
			camel_exception_clear (ex);
		}
		
	} while (!status && id == CAMEL_EXCEPTION_SERVICE_CANT_AUTHENTICATE);
	
	return TRUE;
}

static gboolean
ogo_disconnect (CamelService *service, gboolean clean, CamelException *ex)
{
	CamelOGoStore *store = CAMEL_OGO_STORE (service);
	
	if (!CAMEL_SERVICE_CLASS (parent_class)->disconnect (service, clean, ex))
		return FALSE;
	
	return TRUE;
}

static CamelFolder *
get_folder (CamelStore *store, const char *folder_name, guint32 flags, CamelException *ex)
{
	return NULL;
}

static CamelFolder *
get_trash (CamelStore *store, CamelException *ex)
{
	/* no-op */
	return NULL;
}
