/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Authors :
 *  
 *  Sivaiah Nallagatla <snallagatla@novell.com>
 *
 * Copyright 2003, Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "camel-ogo-listener.h"
#include <string.h>
#include  <camel/camel-i18n.h>
#include <e-util/e-passwords.h>

/*stores some info about all currently existing ogo accounts 
  list of OGoAccountInfo structures */

static 	GList *ogo_accounts = NULL;

struct _CamelOGoListenerPrivate {
	GConfClient *gconf_client;
	/* we get notification about mail account changes form this object */
	EAccountList *account_list;                  
};

typedef struct {
	char *uid;
	char *name;
	char *source_url;
} OGoAccountInfo;

#define OGO_URI_PREFIX   "ogo://" 
#define OGO_PREFIX_LENGTH 6

#define PARENT_TYPE G_TYPE_OBJECT

static GObjectClass *parent_class = NULL;

static void dispose (GObject *object);
static void finalize (GObject *object);


static void 
camel_ogo_listener_class_init (CamelOGoListenerClass *class)
{
	GObjectClass *object_class;
	
	parent_class =  g_type_class_ref (PARENT_TYPE);
	object_class = G_OBJECT_CLASS (class);
	
	/* virtual method override */
	object_class->dispose = dispose;
	object_class->finalize = finalize;
}

static void 
camel_ogo_listener_init (CamelOGoListener *config_listener,  CamelOGoListenerClass *class)
{
	config_listener->priv = g_new0 (CamelOGoListenerPrivate, 1);	
}

static void 
dispose (GObject *object)
{
	CamelOGoListener *config_listener = CAMEL_OGO_LISTENER (object);
	
	g_object_unref (config_listener->priv->gconf_client);
	g_object_unref (config_listener->priv->account_list);

	G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void 
finalize (GObject *object)
{
  	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*determines whehter the passed in account is ogo or not by looking at source url */

static gboolean
get_is_ogo_account (EAccount *account)
{
	if (account->source->url != NULL) {
		return (strncmp (account->source->url,  OGO_URI_PREFIX, OGO_PREFIX_LENGTH ) == 0);
	} else {
		return FALSE;
	}
}

static OGoAccountInfo* 
lookup_account_info (const char *key)
{
	GList *list;
        OGoAccountInfo *info ;
	int found = 0;
                                                                      
        if (!key)
                return NULL;

	info = NULL;

        for (list = g_list_first (ogo_accounts);  list;  list = g_list_next (list)) {
                info = (OGoAccountInfo *) (list->data);
                found = (strcmp (info->uid, key) == 0);
		if (found)
			break;
	}
	if (found)
		return info;
	return NULL;
}

#define CALENDAR_SOURCES "/apps/evolution/calendar/sources"
#define TASKS_SOURCES "/apps/evolution/tasks/sources"
#define SELECTED_CALENDARS "/apps/evolution/calendar/display/selected_calendars"
#define SELECTED_TASKS   "/apps/evolution/calendar/tasks/selected_tasks"

static void 
remove_esource (const char *conf_key, const char *group_name, char* source_name, const char* relative_uri)
{
	ESourceList *list;
        ESourceGroup *group;
        ESource *source;
        GSList *groups;
        GSList *sources;
	gboolean found_group;
	GConfClient* client;
	GSList *ids;
	GSList *node_tobe_deleted;
	char *source_selection_key;
                                                                                                                             
        client = gconf_client_get_default();
        list = e_source_list_new_for_gconf (client, conf_key);
	groups = e_source_list_peek_groups (list); 
	
	found_group = FALSE;
	
	for ( ; groups != NULL && !found_group; groups = g_slist_next (groups)) {

		group = E_SOURCE_GROUP (groups->data);
		
		if (strcmp (e_source_group_peek_name (group), group_name) == 0 && 
		   strcmp (e_source_group_peek_base_uri (group), OGO_URI_PREFIX ) == 0) {

			sources = e_source_group_peek_sources (group);
			
			for( ; sources != NULL; sources = g_slist_next (sources)) {
				
				source = E_SOURCE (sources->data);
				
				if (strcmp (e_source_peek_relative_uri (source), relative_uri) == 0) {
				
					if (!strcmp (conf_key, CALENDAR_SOURCES)) 
						source_selection_key = SELECTED_CALENDARS;
					else if (!strcmp (conf_key, TASKS_SOURCES))
						source_selection_key = SELECTED_TASKS;
					else source_selection_key = NULL;
					if (source_selection_key) {
						ids = gconf_client_get_list (client, source_selection_key , 
									     GCONF_VALUE_STRING, NULL);
						node_tobe_deleted = g_slist_find_custom (ids, e_source_peek_uid (source), (GCompareFunc) strcmp);
						if (node_tobe_deleted) {
							g_free (node_tobe_deleted->data);
							ids = g_slist_delete_link (ids, node_tobe_deleted);
						}
						gconf_client_set_list (client,  source_selection_key, 
								       GCONF_VALUE_STRING, ids, NULL);

					}
					e_source_list_remove_group (list, group);
					e_source_list_sync (list, NULL);	
					found_group = TRUE;
					break;
					
				}
			}

		}
		
	      
	}

	g_object_unref (list);
	g_object_unref (client);		
	
}

static void
add_esource (const char *conf_key, const char *group_name,  const char *source_name, const gchar *item_path, CamelURL *url)
{
	ESourceList *source_list;
	ESourceGroup *group;
	ESource *source;
        GConfClient* client;
	GSList *ids, *temp ;
	char *source_selection_key;
	char *relative_uri;
	
	client = gconf_client_get_default();	
	source_list = e_source_list_new_for_gconf (client, conf_key);

	group = e_source_group_new (group_name,  OGO_URI_PREFIX);
	if (!e_source_list_add_group (source_list, group, -1))
		return;
	relative_uri = g_strdup_printf ("%s@%s/", url->user, url->host);
	
	source = e_source_new (source_name, relative_uri);
	e_source_set_property (source, "auth", "1");
	e_source_set_property (source, "username", url->user);
	e_source_set_property (source, "auth-domain", "OpenGroupware.org");
	e_source_set_property (source, "item_path", item_path);
	e_source_group_add_source (group, source, -1);
	e_source_list_sync (source_list, NULL);

	if (!strcmp (conf_key, CALENDAR_SOURCES)) 
		source_selection_key = SELECTED_CALENDARS;
	else if (!strcmp (conf_key, TASKS_SOURCES))
		source_selection_key = SELECTED_TASKS;
	else source_selection_key = NULL;
	if (source_selection_key) {
		ids = gconf_client_get_list (client, source_selection_key , GCONF_VALUE_STRING, NULL);
		ids = g_slist_append (ids, g_strdup (e_source_peek_uid (source)));
		gconf_client_set_list (client,  source_selection_key, GCONF_VALUE_STRING, ids, NULL);
		temp  = ids;
		for (; temp != NULL; temp = g_slist_next (temp))
			g_free (temp->data);
		g_slist_free (ids);
	}
	
	g_object_unref (source);
	g_object_unref (group);
	g_object_unref (source_list);
	g_object_unref (client);
	g_free (relative_uri);
}


/* looks up for e-source with having same info as old_account_info and changes its values passed in new values */

static void 
modify_esource (const char* conf_key, OGoAccountInfo *old_account_info, const char* new_group_name, CamelURL *new_url)
{
	ESourceList *list;
        ESourceGroup *group;
        ESource *source;
        GSList *groups;
	GSList *sources;
	char *old_relative_uri;
	CamelURL *url;
	gboolean found_group;
      	GConfClient* client;
	char *new_relative_uri;
	
	url = camel_url_new (old_account_info->source_url, NULL);
	
	old_relative_uri =  g_strdup_printf ("%s@%s/", url->user, url->host);
	client = gconf_client_get_default ();
        list = e_source_list_new_for_gconf (client, conf_key);
	groups = e_source_list_peek_groups (list); 

	found_group = FALSE;

	for ( ; groups != NULL &&  !found_group; groups = g_slist_next (groups)) {

		group = E_SOURCE_GROUP (groups->data);
		
		if (strcmp (e_source_group_peek_name (group), old_account_info->name) == 0 && 
		    strcmp (e_source_group_peek_base_uri (group), OGO_URI_PREFIX) == 0) {

			sources = e_source_group_peek_sources (group);
			
			for ( ; sources != NULL; sources = g_slist_next (sources)) {
				
				source = E_SOURCE (sources->data);
				
				if (strcmp (e_source_peek_relative_uri (source), old_relative_uri) == 0) {
					
					new_relative_uri = g_strdup_printf ("%s@%s/", new_url->user, new_url->host);
					e_source_group_set_name (group, new_group_name);
					e_source_set_relative_uri (source, new_relative_uri);
					e_source_set_property (source, "username", new_url->user);
					e_source_list_sync (list, NULL);
					found_group = TRUE;
					g_free (new_relative_uri);
					break;
				}
			}
		}
	}

	g_object_unref (list);
	g_object_unref (client);
	camel_url_free (url);
	g_free (old_relative_uri);

	
}

/* add sources for calendar and tasks if the account added is ogo account
   adds the new account info to  ogo_accounts list */

static void 
add_calendar_tasks_sources (OGoAccountInfo *info)
{
	CamelURL *url;
	char *relative_uri;
	const char *soap_port;
	const char * use_ssl;
	const char *poa_address;
	
	url = camel_url_new (info->source_url, NULL);

	relative_uri =  g_strdup_printf ("%s@%s/", url->user, url->host);
	add_esource ("/apps/evolution/calendar/sources", info->name, _("Calendar"), "public/Calendar", url);
	add_esource ("/apps/evolution/tasks/sources", info->name, _("Tasks"), "Tasks", url);
	
	camel_url_free (url);
	g_free (relative_uri);

}

/* removes calendar and tasks sources if the account removed is ogo account 
   removes the the account info from ogo_account list */

static void 
remove_calendar_tasks_sources (OGoAccountInfo *info)
{
	CamelURL *url;
	char *relative_uri;
        const char *soap_port;
	const char *poa_address;

	url = camel_url_new (info->source_url, NULL);

	relative_uri =  g_strdup_printf ("%s@%s/", url->user, poa_address);
	remove_esource ("/apps/evolution/calendar/sources", info->name, _("Calendar"), relative_uri);
	remove_esource ("/apps/evolution/tasks/sources", info->name,  _("Tasks"), relative_uri);
	camel_url_free (url);
	g_free (relative_uri);

}

typedef struct {
	gchar *name;
	gchar *item_path;
} AddressFolderInfo;

static void
address_folder_info_free (AddressFolderInfo *info)
{
	g_free (info->name);
	g_free (info->item_path);

	g_free (info);
}

static GList*
get_addressbook_names_from_server (char *source_url)
{
	GList *list = NULL;
	AddressFolderInfo *info;
	
	/* Doesn't really get the names from the server */

	info = g_new (AddressFolderInfo, 1);
	info->name = g_strdup (_("Contacts"));
	info->item_path = g_strdup ("public/Contacts");
	
	list = g_list_prepend (list, info);

	return list;
}

static gboolean
add_addressbook_sources (EAccount *account)
{
	CamelURL *url;
	ESourceList *list;
        ESourceGroup *group;
        ESource *source;
       	char *base_uri;
	GList *books_list, *temp_list;
	GConfClient* client;

	g_print ("adding addressbook sources: %s\n", account->source->url);
	
        url = camel_url_new (account->source->url, NULL);
	if (url == NULL) {
		return FALSE;
	}

	base_uri =  g_strdup_printf ("ogo://%s@%s", url->user, url->host);
	client = gconf_client_get_default ();
	list = e_source_list_new_for_gconf (client, "/apps/evolution/addressbook/sources" );
	group = e_source_group_new (account->name, base_uri);
	g_print ("creating new source group: %s\n", account->name, base_uri);
	books_list = get_addressbook_names_from_server (account->source->url);
	temp_list = books_list;
	if (!temp_list)
		return FALSE;
	for (; temp_list != NULL; temp_list = g_list_next (temp_list)) {
		AddressFolderInfo *info = temp_list->data;
		source = e_source_new (info->name, g_strconcat (";",info->item_path, NULL));
		e_source_set_property (source, "auth", "plain/password");
		e_source_set_property (source, "auth-domain", "OpenGroupware.org");
		e_source_set_property(source, "user", url->user);
		e_source_group_add_source (group, source, -1);
		g_object_unref (source);
		address_folder_info_free (info);
		
	}
		
	g_list_free (books_list);			

      
	e_source_list_add_group (list, group, -1);
      	e_source_list_sync (list, NULL);	
	g_object_unref (group);
	g_object_unref (list);
	g_object_unref (client);
	g_free (base_uri);
	
	return TRUE;
}

static void 
modify_addressbook_sources (EAccount *account, OGoAccountInfo *existing_account_info )
{
	CamelURL *url;
	ESourceList *list;
        ESourceGroup *group;
	GSList *groups;
       	gboolean found_group;
	gboolean delete_group;
	char *old_base_uri;
	char *new_base_uri;
	const char *soap_port;
	const char *use_ssl;
	GSList *sources;
	ESource *source;
	GConfClient *client;
	const char *poa_address;
	

	url = camel_url_new (existing_account_info->source_url, NULL);
	if (url == NULL) {
		return;
	}

	old_base_uri =  g_strdup_printf ("ogo://%s@%s", url->user, url->host);
	camel_url_free (url);
	
	url = camel_url_new (account->source->url, NULL);
	if (url == NULL)
		return ;

	new_base_uri = g_strdup_printf ("ogo://%s@%s", url->user, url->host);

	client = gconf_client_get_default ();
	list = e_source_list_new_for_gconf (client, "/apps/evolution/addressbook/sources" );
	groups = e_source_list_peek_groups (list); 
	delete_group = FALSE;
	if (strcmp (old_base_uri, new_base_uri) != 0)
		delete_group = TRUE;
	group = NULL;
	found_group = FALSE;
	for ( ; groups != NULL &&  !found_group; groups = g_slist_next (groups)) {

		group = E_SOURCE_GROUP (groups->data);
		if ( strcmp ( e_source_group_peek_base_uri(group), old_base_uri) == 0 && strcmp (e_source_group_peek_name (group), existing_account_info->name) == 0) {
			found_group = TRUE;
			if (!delete_group) {
				e_source_group_set_name (group, account->name);
				sources = e_source_group_peek_sources (group);
				for (; sources != NULL; sources = g_slist_next (sources)) {
					source = E_SOURCE (sources->data);
					e_source_set_property (source, "use_ssl", use_ssl);
				}
					
				e_source_list_sync (list, NULL);
			}
		
		}
	}
	if (found_group && delete_group) {
		e_source_list_remove_group (list, group);
		e_source_list_sync (list, NULL);
		g_object_unref (list);
		list = NULL;
		add_addressbook_sources (account);
	}
	g_free (old_base_uri);
	if (list)
		g_object_unref (list);
	camel_url_free (url);
	g_object_unref (client);


}

static void 
remove_addressbook_sources (OGoAccountInfo *existing_account_info)
{
	ESourceList *list;
        ESourceGroup *group;
	GSList *groups;
       	gboolean found_group;
	CamelURL *url;
	char *base_uri;
	const char *soap_port;
	GConfClient *client;
	const char *poa_address;

	url = camel_url_new (existing_account_info->source_url, NULL);
	if (url == NULL) {
		return;
	}

	base_uri =  g_strdup_printf ("ogo://%s@%s", url->user,  url->host);
	client = gconf_client_get_default ();
	list = e_source_list_new_for_gconf (client, "/apps/evolution/addressbook/sources" );
	groups = e_source_list_peek_groups (list); 

	found_group = FALSE;

	for ( ; groups != NULL &&  !found_group; groups = g_slist_next (groups)) {

		group = E_SOURCE_GROUP (groups->data);
		if ( strcmp ( e_source_group_peek_base_uri (group), base_uri) == 0 && strcmp (e_source_group_peek_name (group), existing_account_info->name) == 0) {

			e_source_list_remove_group (list, group);
			e_source_list_sync (list, NULL);
			found_group = TRUE;
						
		}
	}
	g_object_unref (list);
	g_object_unref (client);
	g_free (base_uri);
	camel_url_free (url);
	

}

static void 
account_added (EAccountList *account_listener, EAccount *account)
{

	OGoAccountInfo *info;
	gboolean status;

	if (!get_is_ogo_account (account))
		return;
	
	info = g_new0 (OGoAccountInfo, 1);
	info->uid = g_strdup (account->uid);
	info->name = g_strdup (account->name);
	info->source_url = g_strdup (account->source->url);
	status = add_addressbook_sources (account);
	if (status) 
		add_calendar_tasks_sources (info);
	ogo_accounts = g_list_append (ogo_accounts, info);

}

static void 
account_removed (EAccountList *account_listener, EAccount *account)
{
       	OGoAccountInfo *info;
	
	if (!get_is_ogo_account (account))
		return;
	
	info = lookup_account_info (account->uid);
	if (info == NULL) {
		return;
	}

	remove_calendar_tasks_sources (info);
	remove_addressbook_sources (info);
	ogo_accounts = g_list_remove (ogo_accounts, info);
	g_free (info->uid);
	g_free (info->name);
	g_free (info->source_url);
        g_free (info);
}

static void
account_changed (EAccountList *account_listener, EAccount *account)
{
	gboolean is_ogo_account;
	CamelURL *old_url, *new_url;
	const char *old_soap_port, *new_soap_port;
	OGoAccountInfo *existing_account_info;
	const char *old_use_ssl, *new_use_ssl;
	gboolean old_ssl, new_ssl;
	const char *old_poa_address, *new_poa_address;
	
	is_ogo_account = get_is_ogo_account (account);
	
	existing_account_info = lookup_account_info (account->uid);
       
	if (existing_account_info == NULL && is_ogo_account) {

		if (!account->enabled)
			return;

		/* some account of other type is changed to Ogo */
		account_added (account_listener, account);

	} else if ( existing_account_info != NULL && !is_ogo_account) {

		/*Ogo account is changed to some other type */
		remove_calendar_tasks_sources (existing_account_info);
		remove_addressbook_sources (existing_account_info);
		ogo_accounts = g_list_remove (ogo_accounts, existing_account_info);
		g_free (existing_account_info->uid);
		g_free (existing_account_info->name);
		g_free (existing_account_info->source_url);
		g_free (existing_account_info);
		
	} else if ( existing_account_info != NULL && is_ogo_account ) {
		
		if (!account->enabled) {
			account_removed (account_listener, account);
			return;
		}
		old_ssl = new_ssl = FALSE;
		/* some info of ogo account is changed . update the sources with new info if required */
		old_url = camel_url_new (existing_account_info->source_url, NULL);
		new_url = camel_url_new (account->source->url, NULL);

		if (strcmp (old_url->user, new_url->user) != 0) {
			account_removed (account_listener, account);
			account_added (account_listener, account);
		} else if (strcmp (existing_account_info->name, account->name)) {
			
			modify_esource ("/apps/evolution/calendar/sources", existing_account_info, account->name, new_url);
			modify_esource ("/apps/evolution/tasks/sources", existing_account_info, account->name,  new_url);
			modify_addressbook_sources (account, existing_account_info);
			
		}
		
		g_free (existing_account_info->name);
		g_free (existing_account_info->source_url);
		existing_account_info->name = g_strdup (account->name);
		existing_account_info->source_url = g_strdup (account->source->url);
		camel_url_free (old_url);
		camel_url_free (new_url);
	}
		
	g_print ("account changed!\n");
}
      
static void
camel_ogo_listener_construct (CamelOGoListener *config_listener)
{
	EIterator *iter;
	EAccount *account;
	OGoAccountInfo *info ;
	
       	config_listener->priv->account_list = e_account_list_new (config_listener->priv->gconf_client);

	for ( iter = e_list_get_iterator (E_LIST ( config_listener->priv->account_list) ) ; e_iterator_is_valid (iter); e_iterator_next (iter) ) {
		
		account = E_ACCOUNT (e_iterator_get (iter));
		if ( get_is_ogo_account (account) && account->enabled) {
			
		        info = g_new0 (OGoAccountInfo, 1);
			info->uid = g_strdup (account->uid);
			info->name = g_strdup (account->name);
			info->source_url = g_strdup (account->source->url);
			ogo_accounts = g_list_append (ogo_accounts, info);
			
		}

		g_print ("found %d accounts\n", g_list_length (ogo_accounts));
			
	}
	g_signal_connect (config_listener->priv->account_list, "account_added", G_CALLBACK (account_added), NULL);
	g_signal_connect (config_listener->priv->account_list, "account_changed", G_CALLBACK (account_changed), NULL);
	g_signal_connect (config_listener->priv->account_list, "account_removed", G_CALLBACK (account_removed), NULL);
}

GType
camel_ogo_listener_get_type (void)
{
	static GType camel_ogo_listener_type  = 0;
	
	if (!camel_ogo_listener_type) {
		static GTypeInfo info = {
                        sizeof (CamelOGoListenerClass),
                        (GBaseInitFunc) NULL,
                        (GBaseFinalizeFunc) NULL,
                        (GClassInitFunc) camel_ogo_listener_class_init,
                        NULL, NULL,
                        sizeof (CamelOGoListener),
                        0,
                        (GInstanceInitFunc) camel_ogo_listener_init
                };
		camel_ogo_listener_type = g_type_register_static (PARENT_TYPE, "CamelOGoListener", &info, 0);
	}

	return camel_ogo_listener_type;
}

CamelOGoListener*
camel_ogo_listener_new ()
{
	CamelOGoListener *config_listener;
       
	config_listener = g_object_new (CAMEL_TYPE_OGO_LISTENER, NULL);
	config_listener->priv->gconf_client = gconf_client_get_default();
	
	camel_ogo_listener_construct (config_listener);

	g_print ("NEW LISTENER!\n");
	
	return config_listener;
}


