/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* Evolution calendar - generic backend class
 *
 * Copyright (C) 2003 Novell, Inc.
 *
 * Authors: Rodrigo Moya <rodrigo@ximian.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <libebook/e-contact.h>
#include "ogo-book-backend-cache.h"

struct _OGoBookBackendCachePrivate {
	char *uri;
};

/* Property IDs */
enum {
	PROP_0,
	PROP_URI
};

static GObjectClass *parent_class = NULL;

static char *
get_filename_from_uri (const char *uri)
{
	char *mangled_uri, *filename;
	int i;

	/* mangle the URI to not contain invalid characters */
	mangled_uri = g_strdup (uri);
	for (i = 0; i < strlen (mangled_uri); i++) {
		switch (mangled_uri[i]) {
		case ':' :
		case '/' :
			mangled_uri[i] = '_';
		}
	}

	/* generate the file name */
	filename = g_build_filename (g_get_home_dir (), ".evolution/cache/addressbook",
				     mangled_uri, "cache.xml", NULL);

	/* free memory */
	g_free (mangled_uri);

	return filename;
}

static void
ogo_book_backend_cache_set_property (GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
	OGoBookBackendCache *cache;
	OGoBookBackendCachePrivate *priv;
	char *cache_file;

	cache = OGO_BOOK_BACKEND_CACHE (object);
	priv = cache->priv;

	switch (property_id) {
	case PROP_URI :
		cache_file = get_filename_from_uri (g_value_get_string (value));
		if (!cache_file)
			break;

		g_object_set (G_OBJECT (cache), "filename", cache_file, NULL);
		g_free (cache_file);

		if (priv->uri)
			g_free (priv->uri);
		priv->uri = g_value_dup_string (value);
		break;
	default :
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
ogo_book_backend_cache_get_property (GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
	OGoBookBackendCache *cache;
	OGoBookBackendCachePrivate *priv;

	cache = OGO_BOOK_BACKEND_CACHE (object);
	priv = cache->priv;

	switch (property_id) {
	case PROP_URI :
		g_value_set_string (value, priv->uri);
		break;
	default :
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}


static void
ogo_book_backend_cache_finalize (GObject *object)
{
	OGoBookBackendCache *cache;
	OGoBookBackendCachePrivate *priv;

	cache = OGO_BOOK_BACKEND_CACHE (object);
	priv = cache->priv;

	if (priv) {
		if (priv->uri) {
			g_free (priv->uri);
			priv->uri = NULL;
		}

		g_free (priv);
		cache->priv = NULL;
	}

	parent_class->finalize (object);
}

static GObject *
ogo_book_backend_cache_constructor (GType type,
                                 guint n_construct_properties,
                                 GObjectConstructParam *construct_properties)
{
	GObject *obj;
	const char *uri;
	OGoBookBackendCacheClass *klass;
	GObjectClass *parent_class;

	/* Invoke parent constructor. */
	klass = OGO_BOOK_BACKEND_CACHE_CLASS (g_type_class_peek (OGO_TYPE_BOOK_BACKEND_CACHE));
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	obj = parent_class->constructor (type,
					 n_construct_properties,
					 construct_properties);
  
	/* extract uid */
	if (!g_ascii_strcasecmp ( g_param_spec_get_name (construct_properties->pspec), "uri")) {
		uri = g_value_get_string (construct_properties->value);
		g_object_set (obj, "filename", get_filename_from_uri (uri), NULL);
	}

	return obj;
}

static void
ogo_book_backend_cache_class_init (OGoBookBackendCacheClass *klass)
{
	GObjectClass *object_class;

	parent_class = g_type_class_peek_parent (klass);

	object_class = G_OBJECT_CLASS (klass);
	object_class->finalize = ogo_book_backend_cache_finalize;
	object_class->set_property = ogo_book_backend_cache_set_property;
	object_class->get_property = ogo_book_backend_cache_get_property;

        object_class->constructor = ogo_book_backend_cache_constructor;
	g_object_class_install_property (object_class, PROP_URI,
					 g_param_spec_string ("uri", NULL, NULL, "",
							      G_PARAM_READABLE | G_PARAM_WRITABLE
							      | G_PARAM_CONSTRUCT_ONLY));
}

static void
ogo_book_backend_cache_init (OGoBookBackendCache *cache)
{
	OGoBookBackendCachePrivate *priv;

	priv = g_new0 (OGoBookBackendCachePrivate, 1);

	cache->priv = priv;

}

/**
 * ogo_book_backend_cache_get_type:
 * @void:
 *
 * Registers the #OGoBookBackendCache class if necessary, and returns the type ID
 * associated to it.
 *
 * Return value: The type ID of the #OGoBookBackendCache class.
 **/
GType
ogo_book_backend_cache_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static GTypeInfo info = {
                        sizeof (OGoBookBackendCacheClass),
                        (GBaseInitFunc) NULL,
                        (GBaseFinalizeFunc) NULL,
                        (GClassInitFunc) ogo_book_backend_cache_class_init,
                        NULL, NULL,
                        sizeof (OGoBookBackendCache),
                        0,
                        (GInstanceInitFunc) ogo_book_backend_cache_init,
                };
		type = g_type_register_static (E_TYPE_FILE_CACHE, "OGoBookBackendCache", &info, 0);
	}

	return type;
}

/**
 * ogo_book_backend_cache_new
 * @uri: URI of the backend to be cached.
 *
 * Creates a new #OGoBookBackendCache object, which implements a cache of
 * calendar/tasks objects, very useful for remote backends.
 *
 * Return value: The newly created object.
 */
OGoBookBackendCache *
ogo_book_backend_cache_new (const char *uri)
{
	OGoBookBackendCache *cache;
        
       	cache = g_object_new (OGO_TYPE_BOOK_BACKEND_CACHE, "uri", uri, NULL);

        return cache;
}

gboolean
ogo_book_backend_cache_put_contact (OGoBookBackendCache *cache, EContact *contact)
{
	OGoBookBackendCachePrivate *priv;
	const gchar *uid;
	gchar *str;
	gboolean retval;
	
	g_return_val_if_fail (OGO_IS_BOOK_BACKEND_CACHE (cache), FALSE);
	g_return_val_if_fail (E_IS_CONTACT (contact), FALSE);
	
	priv = cache->priv;

	uid = e_contact_get_const (contact, E_CONTACT_UID);

	str = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);

	if (e_file_cache_get_object (E_FILE_CACHE (cache), uid))
		retval = e_file_cache_replace_object (E_FILE_CACHE (cache), uid, str);
	else
		retval = e_file_cache_add_object (E_FILE_CACHE (cache), uid, str);

	return retval;
}

EContact *
ogo_book_backend_cache_get_contact (OGoBookBackendCache *cache, const char *uid)
{
	EContact *contact;
	const gchar *str;
	
	str = e_file_cache_get_object (E_FILE_CACHE (cache), uid);

	if (!str)
		return NULL;

	contact = e_contact_new_from_vcard (str);

	return contact;
}


gboolean
ogo_book_backend_cache_remove_contact (OGoBookBackendCache *cache,
				       const char *uid)
{
	gboolean retval;
	OGoBookBackendCachePrivate *priv;

	g_return_val_if_fail (OGO_IS_BOOK_BACKEND_CACHE (cache), FALSE);
	g_return_val_if_fail (uid != NULL, FALSE);

	priv = cache->priv;

	if (!e_file_cache_get_object (E_FILE_CACHE (cache), uid)) {
		return FALSE;
	}
	
	return e_file_cache_remove_object (E_FILE_CACHE (cache), uid);
}

GList *
ogo_book_backend_cache_get_contacts (OGoBookBackendCache *cache)
{
        char     *comp_str;
        GSList   *l;
	GList    *list = NULL;
	EContact *contact;

        /* return null if cache is not a valid Backend Cache.  */
	g_return_val_if_fail (OGO_IS_BOOK_BACKEND_CACHE (cache), NULL);
	
        l = e_file_cache_get_objects (E_FILE_CACHE (cache));

	for (; l != NULL; l = l->next) {
		/* Don't want the "populated" object */
		if (strcmp (l->data, "TRUE") == 0)
			continue;
		
		contact = e_contact_new_from_vcard (l->data);
		
		if (!contact)
			continue;

		list = g_list_prepend (list, contact);
	}

	return g_list_reverse (list);
}

GSList *
ogo_book_backend_cache_get_keys (OGoBookBackendCache *cache)
{
        /* return null if cache is not a valid Backend Cache.  */
	g_return_val_if_fail (OGO_IS_BOOK_BACKEND_CACHE (cache), NULL);
        return e_file_cache_get_keys (E_FILE_CACHE (cache));
}

void
ogo_book_backend_cache_set_marker (OGoBookBackendCache *cache)
{
	g_return_if_fail (OGO_IS_BOOK_BACKEND_CACHE (cache));
	e_file_cache_add_object (E_FILE_CACHE (cache), "populated", "TRUE");
}

const char *
ogo_book_backend_cache_get_marker (OGoBookBackendCache *cache)
{
	g_return_val_if_fail (OGO_IS_BOOK_BACKEND_CACHE (cache), NULL);
	if (e_file_cache_get_object (E_FILE_CACHE (cache), "populated"))
		return "";
	return NULL;
}
